import { Rule, SubjectType } from './types';

export class Ability {
    private readonly rules: Rule[];

    constructor(rules: Rule[] = []) {
        this.rules = rules;
    }

    can(action: string, subject: SubjectType, context: Record<string, any> = {}): boolean {
        return this.check(action, subject, context, false);
    }

    cannot(action: string, subject: SubjectType, context: Record<string, any> = {}): boolean {
        return !this.can(action, subject, context);
    }

    mustCan(action: string, subject: SubjectType, context: Record<string, any> = {}): void {
        if (!this.can(action, subject, context)) {
            throw new Error(`Permission denied: cannot ${action} ${this._describeSubject(subject)}.`);
        }
    }

    mustCannot(action: string, subject: SubjectType, context: Record<string, any> = {}): void {
        if (!this.cannot(action, subject, context)) {
            throw new Error(`Permission denied: must not be able to ${action} ${this._describeSubject(subject)}.`);
        }
    }

    exportRules(): string {
        return JSON.stringify(this.rules);
    }

    static importRules(json: string): Ability {
        return new Ability(JSON.parse(json));
    }

    private check(action: string, subject: SubjectType, context: Record<string, any>, inverted: boolean): boolean {
        const subjectName = this._getSubjectName(subject);

        for (const rule of this.rules) {
            if (rule.inverted !== inverted) continue;

            const actionMatch = rule.actions.includes(action) || rule.actions.includes('*');
            const subjectMatch = rule.subjects.some(s => this._matchSubject(s, subject, subjectName));

            if (!actionMatch || !subjectMatch) continue;

            if (!rule.conditions) return true;
            if (this._matchConditions(rule.conditions, context)) return true;
        }

        return false;
    }

    private _matchConditions(conditions: Record<string, any>, context: Record<string, any>): boolean {
        return Object.entries(conditions).every(([key, expected]) => {
            const actual = key.split('.').reduce((obj, part) => obj?.[part], context);
            return actual === expected;
        });
    }

    private _getSubjectName(subject: SubjectType): string {
        if (typeof subject === 'string') return subject;
        if (typeof subject === 'function') return subject.name;
        if (typeof subject === 'object' && subject !== null) return subject.constructor.name;
        return '';
    }

    private _matchSubject(ruleSubject: SubjectType, actual: SubjectType, actualName: string): boolean {
        if (ruleSubject === '*') return true;
        if (typeof ruleSubject === 'string') return ruleSubject === actualName;
        if (typeof ruleSubject === 'function') return actual instanceof ruleSubject;
        return ruleSubject === actual;
    }

    private _describeSubject(subject: SubjectType): string {
        if (typeof subject === 'string') return subject;
        if (typeof subject === 'function') return subject.name;
        if (typeof subject === 'object') return subject.constructor.name;
        return 'unknown';
    }
}
